"""
Contains the configuration for the ixoncdkingress webserver
"""

import enum
import json
import os
import sys
from collections.abc import Mapping
from dataclasses import dataclass
from logging import Formatter, Logger, StreamHandler, getLogger
from typing import Any, Final

from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from load_environ_typed import LoaderMap, load
from load_environ_typed.loaders import load_utf8_file_from_path

try:
    from yaml import safe_load as yaml_safe_load
except ImportError:  # pragma: no cover
    yaml_safe_load = None  # type: ignore[assignment]

_LEGACY_CONTEXT_CONFIG_PATH: Final = "./context_config.json"
"""
Path to the (legacy) context config
"""

_JSON_CONTEXT_VALUES_PATH: Final = "./context_values.json"
"""
Path to a JSON file which can be used to load configuration values from.
"""

_YAML_CONTEXT_VALUES_PATH: Final = "./context_values.yaml"
"""
Path to a JSON file which can be used to load configuration values from.
"""


class InvalidConfigException(Exception):
    """
    Exception for invalid values in the config
    """


class WsgiProvider(enum.Enum):
    """
    Provider to use to turn a WSGI application into a (HTTP) server
    """

    WSGIREF = "wsgiref"
    BJOERN = "bjoern"
    WAITRESS = "waitress"
    GUNICORN = "gunicorn"
    CHERRYPY = "cherrypy"
    MEINHELD = "meinheld"


@dataclass
class Config:
    """
    Configuration for the webserver
    """

    production_mode: bool

    document_db_port: int

    http_server_bind: str
    http_server_port: int
    wsgi_provider: WsgiProvider

    logger_log_level: str
    logger_format: str

    cbc_path: str
    api_client_base_url: str

    cbc_signature_public_keys: list[str]

    _logger: Logger | None = None
    _signature_public_keys: list[ed25519.Ed25519PublicKey] | None = None

    @classmethod
    def from_environ(cls, environ: Mapping[str, str]) -> "Config":
        """
        Deprecated, use get_config() instead.

        Parses an environment dictionary and generates a config based on it
        """
        return get_config(environ)

    def get_logger(self) -> Logger:
        """
        Returns a logger
        """
        if self._logger is None:
            logger = getLogger("ixoncdkingress")
            logger.setLevel(self.logger_log_level)

            handler = StreamHandler(sys.stdout)
            handler.setFormatter(Formatter(self.logger_format))
            logger.addHandler(handler)

            self._logger = logger

        return self._logger

    def get_signature_public_keys(self) -> list[ed25519.Ed25519PublicKey]:
        """
        Returns a cached instance of signature public keys
        """
        if self._signature_public_keys is None:
            keys: list[ed25519.Ed25519PublicKey] = []

            for index, pem in enumerate(self.cbc_signature_public_keys):
                key = load_pem_public_key(pem.encode("ASCII"), None)
                assert isinstance(key, ed25519.Ed25519PublicKey), (
                    f"signature_public_key {index + 1} is not a valid Ed25519 private key"
                )
                keys.append(key)

            self._signature_public_keys = keys
        return self._signature_public_keys


def load_str_file(file_path: str | None, default: str) -> str:
    """
    Loads a string from a file, returns default if no path is given
    """
    if not file_path:
        return default

    try:
        with open(file_path, encoding="utf-8") as file:
            string = file.read().rstrip()
            if not string:
                return default
        return string
    except OSError:
        return default


def load_json_file(
    file_path: str,
) -> dict[str, Any]:
    """
    Loads a Dict from a JSON file.
    """
    with open(file_path, encoding="utf-8") as file:
        result = json.load(file)

    if not isinstance(result, dict):
        raise ValueError("File must contain a dictionary")

    return result


def _load_yaml_file(
    file_path: str,
) -> dict[str, Any]:
    """
    Loads a Dict from a yaml file.
    """
    assert yaml_safe_load is not None, "PyYAML is not installed"

    with open(file_path, encoding="utf-8") as file:
        result = yaml_safe_load(file)

    if not isinstance(result, dict):
        raise ValueError("File must contain a dictionary")

    return result


def load_context_values(config: Config) -> dict[str, Any]:
    """
    Load context values from the config file(s).

    First try the YAML file, if it doesn't exist, then try to load the context values from the
    JSON file.
    If both YAML and JSON files don't exist, it will try to use the legacy context config JSON file
    and uses the values form the 'values' field, whereafter it also writes the values to the
    JSON file.
    """
    logger = config.get_logger()

    if os.path.exists(_YAML_CONTEXT_VALUES_PATH):
        if yaml_safe_load is None:
            raise ImportError(
                "Please install PyYAML to use context values from %s",
                _YAML_CONTEXT_VALUES_PATH,
            )
        logger.info("Loading context values from %s", _YAML_CONTEXT_VALUES_PATH)
        return _load_yaml_file(_YAML_CONTEXT_VALUES_PATH)

    if os.path.exists(_JSON_CONTEXT_VALUES_PATH):
        logger.info("Loading context values from %s", _JSON_CONTEXT_VALUES_PATH)
        logger.warning(
            "The file %s will be deprecated, please use %s instead. "
            "YAML is more readable and supports comments which can help during "
            "the configuration in the IXON Cloud",
            _JSON_CONTEXT_VALUES_PATH,
            _YAML_CONTEXT_VALUES_PATH,
        )
        if os.path.exists(_LEGACY_CONTEXT_CONFIG_PATH):
            logger.info(
                "The file %s is unused and can be safely removed",
                _LEGACY_CONTEXT_CONFIG_PATH,
            )

        return load_json_file(_JSON_CONTEXT_VALUES_PATH)

    if os.path.exists(_LEGACY_CONTEXT_CONFIG_PATH):
        logger.info("Loading context config from %s", _LEGACY_CONTEXT_CONFIG_PATH)

        context_config = load_json_file(_LEGACY_CONTEXT_CONFIG_PATH)
        context_values = context_config.get("values")

        if isinstance(context_values, dict):
            logger.warning("The file %s is deprecated", _LEGACY_CONTEXT_CONFIG_PATH)
            logger.warning(
                "Contents of 'values' field from %s copied to %s.",
                _LEGACY_CONTEXT_CONFIG_PATH,
                _JSON_CONTEXT_VALUES_PATH,
            )
            logger.warning(
                "The file %s will no longer be used and can be safely removed.",
                _LEGACY_CONTEXT_CONFIG_PATH,
            )
            with open(_JSON_CONTEXT_VALUES_PATH, "w") as f:
                json.dump(context_values, f)

            return context_values

    return {}


def load_comma_separated_strings(raw: str) -> list[str]:
    return raw.split(",") if raw != "" else []


CONFIG_DEFAULTS = {
    "API_CLIENT_BASE_URL": "https://api.ayayot.com/",
    "CBC_PATH": "/user_scripts",
    "CBC_SIGNATURE_PUBLIC_KEYS": "",
    "DOCUMENT_DB_PORT": "27017",
    "HTTP_SERVER_BIND": "127.0.0.1",
    "HTTP_SERVER_PORT": "8020",
    "LOGGER_FORMAT": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    "LOGGER_LOG_LEVEL": "INFO",
    "PRODUCTION_MODE": "False",
    "WSGI_PROVIDER": "wsgiref",
}

CONFIG_LOADERS: LoaderMap = {
    "cbc_signature_public_keys": load_comma_separated_strings,
    "version": load_utf8_file_from_path,
    "wsgi_provider": WsgiProvider,
}


def get_config(environ: Mapping[str, str] | None = None) -> Config:
    return load(Config, environ=environ, defaults=CONFIG_DEFAULTS, loaders=CONFIG_LOADERS)



import warnings

warnings.warn(
    "ixoncdkingress.cbc has been deprecated, please use ixoncdkingress.function",
    DeprecationWarning,
)

from ixoncdkingress.function.context import (  # noqa: E402
    FunctionContext,
    FunctionResource,
)

CbcContext = FunctionContext
CbcResource = FunctionResource



"""
Context.
"""

import textwrap
import warnings
from collections.abc import Callable
from functools import wraps
from inspect import _empty, signature
from typing import (
    Any,
    Literal,
    overload,
)

from pymongo import MongoClient

from ixoncdkingress.function.api_client import ApiClient
from ixoncdkingress.function.document_db_client import (
    DocumentDBClient,
    DocumentType,
)
from ixoncdkingress.function.time_series_client import TimeSeriesClient
from ixoncdkingress.schema_utils import generate_input_schema, generate_output_schema
from ixoncdkingress.types import (
    ObjectJsonSchema,
    Tool,
    ToolIssue,
)


class FunctionResource:
    """
    Describes an IXAPI resource.
    """

    public_id: str
    name: str
    custom_properties: dict[str, Any]
    permissions: set[str] | None

    def __init__(
        self,
        public_id: str,
        name: str,
        custom_properties: dict[str, Any],
        permissions: set[str] | None,
    ) -> None:
        self.public_id = public_id
        self.name = name
        self.custom_properties = custom_properties
        self.permissions = permissions

    def __repr__(self) -> str:
        return (
            "<FunctionResource"
            f" public_id={self.public_id},"
            f" name={self.name},"
            f" custom_properties={self.custom_properties!r},"
            f" permissions={self.permissions!r},"
            f">"
        )


def get_tool_definition(
    function: Callable[..., Any],
    requires_human_permission: bool = False,
    description: str | None = None,
    examples: list[str] | None = None,
    input_schema: ObjectJsonSchema | None = None,
    output_schema: ObjectJsonSchema | None = None,
    meta: dict[str, Any] | None = None,
    exclude_param_types: set[type] | None = None,
    exclude_param_names: set[str] | None = None,
) -> Tool:
    """
    Get the tool definition for a function.
    """
    issues: list[ToolIssue] = []

    _description = function.__doc__ if description is None else description
    if _description is None:
        issues.append(
            ToolIssue(
                type="description_missing",
                error="Agentic methods require description for docstrings",
            )
        )
    else:
        _description = textwrap.dedent(_description).strip(" \n")

    if exclude_param_types is None:
        exclude_param_types = {FunctionContext}
    else:
        exclude_param_types.add(FunctionContext)

    _input_schema = generate_input_schema(
        function,
        exclude_param_types=exclude_param_types,
        exclude_param_names=exclude_param_names,
    )
    _output_schema = generate_output_schema(function)
    if input_schema is not None:
        _input_schema.schema = input_schema
    elif _input_schema.issue is not None:
        issues.append(_input_schema.issue)

    if output_schema is not None:
        _output_schema.schema = output_schema
    elif _output_schema.issue is not None:
        issues.append(_output_schema.issue)

    return Tool(
        name=function.__name__,
        description=_description,
        requires_human_permission=requires_human_permission,
        examples=examples,
        input_schema=_input_schema.schema,
        output_schema=_output_schema.schema,
        meta=meta,
        issues=issues,
        input_model=_input_schema.model,
        output_model=_output_schema.model,
    )


class FunctionContext:
    """
    The context for a Cloud Function.
    """

    config: dict[str, str]
    api_client: ApiClient
    user: FunctionResource | None
    company: FunctionResource | None
    asset: FunctionResource | None
    agent: FunctionResource | None
    template: FunctionResource | None

    _mongo_client: MongoClient[DocumentType] | None
    _ddb_collection_name: str

    @property
    def document_db_client(self) -> DocumentDBClient | None:
        """
        The client for accessing the default collection of the Document DB.

        This may be None when running the Cloud Function in development mode,
        and using the local debug form without specifying a company.
        """
        if not self._mongo_client or not self.company:
            return None

        return DocumentDBClient(
            self._mongo_client[self.company.public_id][self._ddb_collection_name]
        )

    @property
    def time_series_client(self) -> TimeSeriesClient | None:
        """
        The client for accessing the time series collection of the Cloud Function.

        WARNING: This function and client is experimental and may change in the future.

        This may be None when running the Cloud Function in development mode,
        and using the local debug form without specifying a company.
        """
        warnings.warn(
            "The TimeSeriesClient is experimental and may change in the future",
            category=FutureWarning,
        )

        if not self._mongo_client or not self.company:
            return None

        return TimeSeriesClient(
            self._mongo_client[self.company.public_id][f"{self._ddb_collection_name}_ts"]
        )

    @property
    def agent_or_asset(self) -> FunctionResource:
        """
        Return either an Agent or an Asset resource, depending on what's available. If both are
        available in the context, returns the Asset resource.
        """
        if self.asset:
            return self.asset

        assert self.agent
        return self.agent

    def __init__(
        self,
        config: dict[str, str],
        api_client: ApiClient,
        mongo_client: MongoClient[DocumentType] | None,
        user: FunctionResource | None,
        company: FunctionResource | None,
        asset: FunctionResource | None,
        agent: FunctionResource | None,
        template: FunctionResource | None,
        document_db_collection_name: str,
    ) -> None:
        self.config = config
        self.api_client = api_client
        self.user = user
        self.company = company
        self.asset = asset
        self.agent = agent
        self.template = template

        self._mongo_client = mongo_client
        self._ddb_collection_name = document_db_collection_name

    def __repr__(self) -> str:
        return (
            f"<FunctionContext"
            f" config={self.config!r},"
            f" api_client={self.api_client!r},"
            f" document_db_client={self.document_db_client!r},"
            f" user={self.user!r},"
            f" company={self.company!r},"
            f" asset={self.asset!r},"
            f" agent={self.agent!r},"
            f" template={self.template!r},"
            f">"
        )

    @staticmethod
    def expose_cloud_function(function: Any) -> Any:
        """
        Decorator to mark a function as an exposed endpoint.
        """
        sig = signature(function, eval_str=True)

        if not sig.parameters:
            raise Exception("Function has no argument for FunctionContext")

        # If the first function argument has a type annotation it should be of FunctionContext
        context_param = sig.parameters[next(iter(sig.parameters))]
        if (
            context_param.annotation is not _empty
            and context_param.annotation is not FunctionContext
        ):
            raise Exception("First function parameter should be of type FunctionContext")

        function.exposed = True

        return function

    @overload
    @staticmethod
    def expose(
        function: Callable[..., Any],
        *,
        is_agentic: Literal[False] = False,
        requires_human_permission: bool = False,
        description: str | None = None,
        examples: list[str] | None = None,
        input_schema: ObjectJsonSchema | None = None,
        output_schema: ObjectJsonSchema | None = None,
        meta: dict[str, Any] | None = None,
        exclude_param_types: set[type] | None = None,
        exclude_param_names: set[str] | None = None,
    ) -> Callable[..., Any]: ...

    @overload
    @staticmethod
    def expose(  # pragma: no cover
        function: None = None,
        *,
        is_agentic: Literal[True] = True,
        requires_human_permission: bool = False,
        description: str | None = None,
        examples: list[str] | None = None,
        input_schema: ObjectJsonSchema | None = None,
        output_schema: ObjectJsonSchema | None = None,
        meta: dict[str, Any] | None = None,
        exclude_param_types: set[type] | None = None,
        exclude_param_names: set[str] | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...

    @staticmethod
    def expose(  # pragma: no cover
        function: Callable[..., Any] | None = None,
        *,
        is_agentic: bool = False,
        requires_human_permission: bool = False,
        description: str | None = None,
        examples: list[str] | None = None,
        input_schema: ObjectJsonSchema | None = None,
        output_schema: ObjectJsonSchema | None = None,
        meta: dict[str, Any] | None = None,
        exclude_param_types: set[type] | None = None,
        exclude_param_names: set[str] | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]] | Callable[..., Any]:
        """
        Decorator to mark a function as an exposed tool to an agent.
        Best approach is to use pydantic models for input and output schema, and let this
        decorator generate the schema for you. In case you do not want to use pydantic models,
        and you have non standard python types, you need to provide the input and output
        schema manually.
        """

        def decorator(inner_function: Any) -> Any:
            FunctionContext.expose_cloud_function(inner_function)
            wraps(inner_function)

            if not is_agentic:
                return inner_function

            tool_definition = get_tool_definition(
                function=inner_function,
                description=description,
                examples=examples,
                input_schema=input_schema,
                output_schema=output_schema,
                meta=meta,
                requires_human_permission=requires_human_permission,
                exclude_param_types=exclude_param_types,
                exclude_param_names=exclude_param_names,
            )
            inner_function.tool_definition = tool_definition
            inner_function.exposed_to_agent = is_agentic
            inner_function.input_model = tool_definition.input_model
            inner_function.output_model = tool_definition.output_model

            return inner_function

        return decorator if function is None else decorator(function)



"""
Functionality shared between all sorts of collections
"""

from collections.abc import Mapping
from typing import Any, Generic, TypeVar

from pymongo.collection import Collection as MongoCollection
from pymongo.cursor import Cursor

_T = TypeVar("_T", bound=Mapping[str, object])


class Collection(Generic[_T]):
    """
    A read-only collection of data.
    """

    _collection: MongoCollection[_T]

    def __init__(
        self,
        collection: MongoCollection[_T],
    ) -> None:
        self._collection = collection

    def find_one(self, *args: Any, filter_map: Any | None, **kwargs: Any) -> _T | None:
        """
        Get a single item from the collection.
        """
        return self._collection.find_one(filter_map, *args, **kwargs)

    def find(self, *args: Any, **kwargs: Any) -> Cursor[_T]:
        """
        Query the collection.
        """
        return self._collection.find(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} company={self._collection.database.name},>"



"""
Module containing type aliases and enums
"""

import enum
from dataclasses import dataclass, field
from typing import Any, ClassVar, Literal, TypeAlias, TypedDict

from pydantic import BaseModel

FunctionLocation = tuple[str, str]
"""
Location of the function. First element specifies the Python module path, second element the
Python function name.
"""

FunctionArguments = dict[str, str]
"""Arguments to the function."""


class ContentType(enum.Enum):
    """
    HTTP content types which are used in the Content-Type headers
    """

    FORM = "application/x-www-form-urlencoded"
    HTML = "text/html"
    JSON = "application/json"


class RequestMethod(enum.Enum):
    """
    Enum with valid HTTP methods
    """

    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    DELETE = "DELETE"
    PUT = "PUT"
    OPTIONS = "OPTIONS"


class ResponseCode(enum.Enum):
    """
    HTTP response codes to return
    """

    OK = 200
    NO_CONTENT = 204
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    METHOD_NOT_ALLOWED = 405
    INTERNAL_ERROR = 500
    NOT_AVAILABLE = 503

    def status_line(self) -> str:
        """Produces a HTTP status line: <status-code> <message>"""
        return f"{self.value} {self.name.replace('_', ' ')}"


ObjectJsonSchema: TypeAlias = dict[str, Any]
"""Type representing JSON schema of an object, e.g. where `"type": "object"`.

This type is used to define agents tools input and output schema.

With PEP-728 this should be a TypedDict with `type: Literal['object']`, and `extra_parts=Any`
"""


class ModuleLoadIssue(TypedDict, total=True):
    """Definition for a module load issue."""

    type: str
    """The type of the issue."""
    error: str
    """The error message."""
    module_name: str
    """The name of the module."""


class ToolIssue(TypedDict, total=True):
    """Definition for a tool issue."""

    type: str
    """The type of the issue."""
    error: str
    """The error message."""


@dataclass(kw_only=True)
class Schema:
    """Definition for input or output schema of a tool"""

    schema: ObjectJsonSchema | None = field(default=None)
    """The JSON Schema object defining the expected parameters for the tool."""
    model: type[BaseModel] | None = field(default=None)
    """The Pydantic model for the input or output schema."""
    issue: ToolIssue | None = field(default=None)
    """The issue on schema generation."""


class ToolDefinition(TypedDict):
    """API Tool Definition for the Agentic Tools API"""

    name: str
    """The name of the tool."""
    description: str
    """A human-readable description of the tool."""
    requires_human_permission: bool
    """Whether the tool requires human permission to be called."""
    inputSchema: ObjectJsonSchema | None
    """A JSON Schema object defining the expected parameters for the tool."""
    outputSchema: ObjectJsonSchema | None
    """An optional JSON Schema object defining the structure of the tool's output returned in
    the structuredContent field of a CallToolResult."""
    meta: dict[str, Any] | None
    """Metadata for the tool."""
    issues: list[ToolIssue] | None
    """The issues with the tool."""


@dataclass(kw_only=True)
class Tool:
    """Definition for a tool the client can call."""

    name: str
    """The name of the tool."""
    description: str | None = field(default=None)
    """A human-readable description of the tool."""
    requires_human_permission: bool = field(default=False)
    """Whether the tool requires human permission to be called."""
    examples: list[str] | None = field(default=None)
    """Examples of how to use the tool."""
    input_schema: ObjectJsonSchema | None = field(default=None)
    """A JSON Schema object defining the expected parameters for the tool."""
    output_schema: ObjectJsonSchema | None = field(default=None)
    """
    An optional JSON Schema object defining the structure of the tool's output
    returned in the structuredContent field of a CallToolResult.
    """
    meta: dict[str, Any] | None = field(default=None)
    """
    metadata
    """

    module_name: str | None = field(default=None)
    """The name of the module the tool is defined in."""

    issues: list[ToolIssue] | None = field(default=None)
    """The issues with the tool."""
    input_model: type[BaseModel] | None = field(default=None)
    """The Pydantic model class for the tool input."""
    output_model: type[BaseModel] | None = field(default=None)
    """The Pydantic model class for the tool output."""

    @property
    def is_ready(self) -> bool:
        """Check if the tool is ready for AI agent."""
        return self.description is not None and self.input_schema is not None

    def to_dict(self) -> ToolDefinition:
        """
        Convert the tool to a dictionary.
        """

        examples = "examples:\n" + "\n".join(self.examples) if self.examples else ""

        if self.examples is not None and self.description is not None:
            description = self.description + "\n" + examples
        elif self.examples is not None and self.description is None:
            description = examples
        else:
            description = self.description or ""

        return ToolDefinition(
            name=self.module_name + "." + self.name if self.module_name else self.name,
            description=description,
            requires_human_permission=self.requires_human_permission,
            inputSchema=self.input_schema,
            outputSchema=self.output_schema,
            meta=self.meta,
            issues=self.issues,
        )


class ListToolsResponse(TypedDict, total=True):
    """
    The expected type for a response for the
    List Agent Tools API
    """

    result: Literal["success"]
    data: list[ToolDefinition]


class ListToolsIssuesResponse(TypedDict, total=True):
    """
    The expected type for a response for the
    List Agent Tools Issues API
    """

    result: Literal["success"]
    data: list[ToolDefinition | ModuleLoadIssue]


class AppDescription:
    """App Description class"""

    _instance: "ClassVar[AppDescription | None]" = None

    description: str | None = None
    """A human-readable description of the app."""
    short_name: str | None = None
    """The short name of the app."""

    def __new__(
        cls, description: str | None = None, short_name: str | None = None
    ) -> "AppDescription":
        """Create the app description singleton."""
        if cls._instance is not None and (
            cls._instance.description != description or cls._instance.short_name != short_name
        ):
            raise ValueError("App description already created")
        cls._instance = super().__new__(cls)
        cls._instance.description = description
        cls._instance.short_name = short_name

        return cls._instance

    @classmethod
    def get(cls) -> "AppDescription":
        """Get the app description singleton."""
        if cls._instance is None:
            raise ValueError("App description not set")
        return cls._instance


class AppDescriptionData(TypedDict, total=True):
    """App Description"""

    description: str | None
    shortName: str | None


class AppDescriptionResponse(TypedDict, total=True):
    """App Description"""

    result: Literal["success"]
    data: AppDescriptionData



import warnings

warnings.warn(
    "ixoncdkingress.cbc has been deprecated, please use ixoncdkingress.function",
    DeprecationWarning,
)

from ixoncdkingress.function.document_db_client import (  # noqa: E402, I001
    DocumentDBAuthentication,
    DocumentDBClient,
    DocumentType,
    TIMEOUT,
)

__all__ = ["TIMEOUT", "DocumentDBAuthentication", "DocumentDBClient", "DocumentType"]



from collections.abc import Callable
from inspect import _empty, signature
from typing import Any, get_type_hints

from pydantic import (
    Field,
    PydanticSchemaGenerationError,
    create_model,
)

from ixoncdkingress.types import Schema, ToolIssue

# ---------------------------------------------------------------------------
# Parameter collection
# ---------------------------------------------------------------------------


def _build_input_fields(
    sig: Any,
    type_hints: dict[str, Any],
    *,
    exclude_param_types: set[type] | None = None,
    exclude_param_names: set[str] | None = None,
) -> dict[str, Any]:
    """Return ``{name: (hint, default)}`` for non-excluded parameters."""
    if exclude_param_types is None:
        exclude_param_types = set()

    if exclude_param_names is None:
        exclude_param_names = set()

    fields: dict[str, Any] = {}
    for name, param in sig.parameters.items():
        if param.annotation is _empty:
            continue
        hint = type_hints.get(name)
        if hint in exclude_param_types or name in exclude_param_names or name.startswith("_"):
            continue
        default = param.default if param.default is not _empty else ...
        fields[name] = (hint, default)
    return fields


# ---------------------------------------------------------------------------
# Public API
# ---------------------------------------------------------------------------


def generate_input_schema(
    function: Callable[..., Any],
    *,
    exclude_param_types: set[type] | None = None,
    exclude_param_names: set[str] | None = None,
) -> Schema:
    """Generate the input schema for a function.

    Uses pydantic's `create_model` / `model_json_schema` for the base schema.

    Annotated aliases with `Field(json_schema_extra={"definition": ...})` are propagated natively
    by pydantic for all positions(union members, list/set/tuple items, dict keys) via
    `propertyNames` and for dict values this via `additionalProperties`.
    """
    sig = signature(function)
    # include_extras=True preserves Annotated metadata so pydantic can merge
    # json_schema_extra into the generated schema nodes.
    type_hints = get_type_hints(function, include_extras=True)
    merged_exclude_names: set[str] = set(exclude_param_names) if exclude_param_names else set()

    fields = _build_input_fields(
        sig,
        type_hints,
        exclude_param_types=exclude_param_types,
        exclude_param_names=merged_exclude_names,
    )
    try:
        input_model = create_model("InputSchema", **fields)
        input_schema = input_model.model_json_schema()
        input_schema.pop("title", None)
        return Schema(
            schema=input_schema,
            model=input_model,
        )
    except PydanticSchemaGenerationError as e:
        return Schema(
            issue=ToolIssue(
                type="input_schema_generation_error",
                error=str(e),
            ),
        )


def generate_output_schema(function: Callable[..., Any]) -> Schema:
    """
    Generate the output schema for a function.

    Uses pydantic's `create_model` / `model_json_schema` for the base schema.

    Annotated aliases with `Field(json_schema_extra={"definition": ...})` are propagated natively
    by pydantic for all positions(union members, list/set/tuple items, dict keys) via
    `propertyNames` and for dict values this via `additionalProperties`.
    """
    sig = signature(function)
    # include_extras=True preserves Annotated metadata so pydantic can merge
    # json_schema_extra into the generated schema nodes.
    type_hints = get_type_hints(function, include_extras=True)

    if sig.return_annotation is _empty:
        return Schema()
    try:
        output_model = create_model(
            "OutputSchema",
            output=(type_hints["return"], Field(default=...)),
        )
        output_schema = output_model.model_json_schema()
        output_schema.pop("title")
        return Schema(
            schema=output_schema,
            model=output_model,
        )

    except PydanticSchemaGenerationError as e:
        return Schema(
            issue=ToolIssue(
                type="output_schema_generation_error",
                error=str(e),
            ),
        )



import warnings

warnings.warn(
    "ixoncdkingress.cbc has been deprecated, please use ixoncdkingress.function",
    DeprecationWarning,
)

from ixoncdkingress.function.objectstorage.types import (  # noqa: E402
    ListPathResponse,
    PathData,
    PathMapping,
    PathResponse,
    ResourceType,
)

__all__ = [
    "ListPathResponse",
    "PathData",
    "PathMapping",
    "PathResponse",
    "ResourceType",
]



import warnings

warnings.warn(
    "ixoncdkingress.cbc has been deprecated, please use ixoncdkingress.function",
    DeprecationWarning,
)

from ixoncdkingress.function.api_client import ApiClient  # noqa: E402

__all__ = ["ApiClient"]



"""Cloud Function servlet module."""

import json
from datetime import datetime

from cryptography.exceptions import InvalidSignature

from ixoncdkingress.function.caller import call_function
from ixoncdkingress.function.context import FunctionContext
from ixoncdkingress.types import ContentType, FunctionArguments, FunctionLocation, ResponseCode
from ixoncdkingress.utils import handle_exception
from ixoncdkingress.webserver.config import Config, load_context_values
from ixoncdkingress.webserver.form import generate_form, parse_form_input
from ixoncdkingress.webserver.request import Request
from ixoncdkingress.webserver.response import Response
from ixoncdkingress.webserver.utils import parse_json_input, read_qs_as_dict


class Servlet:
    """Servlet handling Cloud Function calls."""

    config: Config

    def __init__(self, config: Config) -> None:
        self.config = config

    def do_options(self, request: Request, response: Response) -> None:
        """
        Handle an OPTIONS request
        """
        del request

        response.status_code = ResponseCode.NO_CONTENT
        response.content_type = ContentType.HTML
        response.headers = [
            # Only useful for testing with swagger
            # when not running behind a production nginx
            ("Access-Control-Allow-Credentials", "false"),
            ("Access-Control-Allow-Headers", "*"),
            ("Access-Control-Allow-Methods", "*"),
            ("Access-Control-Allow-Origin", "*"),
            ("Access-Control-Expose-Headers", "*"),
            ("Access-Control-Max-Age", "30"),  # cache time of the above
        ]

    def do_get(self, request: Request, response: Response) -> None:
        """
        Handle a GET request
        """
        response.status_code = ResponseCode.OK
        response.headers = [("Content-Type", ContentType.HTML.value)]

        pre_fill = {}

        if request.cookies:
            pre_fill = {k: v.value for k, v in request.cookies.items()}

        response.set_body(bytes(generate_form(pre_fill), "utf-8"))

    def do_post(self, request: Request, response: Response) -> None:
        """
        Handle a POST request
        """
        available_content_types = [ContentType.JSON]

        context_values = {}

        if not self.config.production_mode:
            # Only JSON requests are allowed in production mode
            available_content_types.append(ContentType.FORM)
            try:
                context_values = load_context_values(self.config)
            except json.JSONDecodeError as exception:
                handle_exception(exception, response)
                return

            logger = self.config.get_logger()
            logger.info("context values: %s", context_values)

        if request.content_type not in available_content_types:
            raise NotImplementedError

        context, function_location, function_arguments, created_on = self._parse_body(
            request.request_body, request.content_type, context_values
        )

        if self.config.production_mode:
            self._verify_request(request, response, created_on)

            if response.status_code == ResponseCode.UNAUTHORIZED:
                return

        out_put = call_function(
            self.config, context, function_location, function_arguments, response
        )

        if response.status_code == ResponseCode.INTERNAL_ERROR:
            return

        if ContentType.FORM == request.content_type:
            pre_fill = read_qs_as_dict(request.request_body)
            response.set_body(
                bytes(generate_form(pre_fill, json.dumps(out_put, indent=4)), "utf-8")
            )
            response.content_type = ContentType.HTML
            response.cookie = {k: f"{v}; Max-Age=28800" for k, v in pre_fill.items()}
        else:  # ContentType.JSON by exclusion
            response.set_body(bytes(json.dumps(out_put), "utf-8"))
            response.content_type = request.content_type
            response.headers.append(("Access-Control-Allow-Origin", "*"))

    def _parse_body(
        self,
        in_put: bytes,
        content_type: ContentType,
        context_config: dict[str, str],
    ) -> tuple[FunctionContext, FunctionLocation, FunctionArguments, str]:
        """
        Parses the request body to a key-value dictionary.
        """
        body = in_put.decode("utf-8")

        if content_type == ContentType.FORM:
            return parse_form_input(self.config, context_config, body)

        if content_type == ContentType.JSON:
            return parse_json_input(self.config, context_config, body)

        raise NotImplementedError

    def _verify_request(self, request: Request, response: Response, created_on: str | None) -> None:
        if not created_on or not request.authorization:
            response.status_code = ResponseCode.UNAUTHORIZED
            return

        request_datetime = datetime.fromisoformat(created_on)
        difference = datetime.now(tz=request_datetime.tzinfo) - request_datetime

        one_minute_secs = 60
        if difference.total_seconds() > one_minute_secs:
            response.status_code = ResponseCode.UNAUTHORIZED
            return

        if " " not in request.authorization:
            response.status_code = ResponseCode.UNAUTHORIZED
            return

        scheme, signature = request.authorization.split(" ", 1)

        if scheme != "signature":
            response.status_code = ResponseCode.UNAUTHORIZED
            return

        signature_public_keys = self.config.get_signature_public_keys()

        request_signature_valid = False
        for key in signature_public_keys:
            try:
                key.verify(bytes.fromhex(signature), request.request_body)
                request_signature_valid = True
            except InvalidSignature:  # noqa: PERF203
                continue

        if not request_signature_valid:
            response.status_code = ResponseCode.UNAUTHORIZED



"""
Types that can be used when implementing
the Object Storage interface.

Compatible with ayayot.objectstorage_v1
"""

import enum
from typing import Literal, TypedDict


class ResourceType(str, enum.Enum):
    """
    A type of resource for which a path is returned
    """

    ASSET = "Asset"
    AGENT = "Agent"


class PathData(TypedDict, total=True):
    """
    Response data for a single path
    """

    path: str


class PathMapping(TypedDict, total=True):
    """
    A mapping between a certain resource and a path,
    as returned by authorize_list
    """

    publicId: str | None
    path: str
    type: ResourceType


class PathResponse(TypedDict, total=True):
    """
    The expected type for a response for the
    Object Storage authorize_upload, authorize_download &
    authorize_delete APIs
    """

    result: Literal["success"]
    data: PathData


class ListPathResponse(TypedDict, total=True):
    """
    The expected type for a response for the
    Object Storage authorize_list API
    """

    result: Literal["success"]
    data: list[PathMapping]



"""
WGSI implementation for ixoncdkingress
"""

import functools
import signal
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from ixoncdkingress.types import RequestMethod, ResponseCode
from ixoncdkingress.webserver.config import Config, WsgiProvider
from ixoncdkingress.webserver.request import Request
from ixoncdkingress.webserver.response import Response
from ixoncdkingress.webserver.servlet import Servlet
from ixoncdkingress.webserver.utils import setup_mongo_container, shutdown_mongo_container

if TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment


def restapi_wsgi(
    config: Config,
    servlet: Servlet,
    environ: "WSGIEnvironment",
    start_response: "StartResponse",
) -> Iterable[bytes]:
    """
    Main application to give to a WSGI-capable (HTTP) server
    """
    request = Request(config, environ)
    response = Response(start_response)

    try:
        do_servlet_method(servlet, request, response)
    except NotImplementedError:
        return start_error_response(start_response)

    return response()


def do_servlet_method(servlet: Servlet, request: Request, response: Response) -> None:
    """
    Calls the Servlet method specified by the `request.request_method`
    """
    if RequestMethod.OPTIONS == request.request_method:
        servlet.do_options(request, response)
    elif RequestMethod.GET == request.request_method and not servlet.config.production_mode:
        # Get requests are only allowed in development (non production mode)
        servlet.do_get(request, response)
    elif RequestMethod.POST == request.request_method:
        servlet.do_post(request, response)
    else:
        raise NotImplementedError


def start_error_response(
    start_response: "StartResponse",
) -> Iterable[bytes]:
    """
    Creates a new Response producing an error response
    """
    error_response = Response(start_response)
    error_response.status_code = ResponseCode.METHOD_NOT_ALLOWED
    return error_response()


def run_server(config: Config, servlet: Servlet) -> None:
    """
    Runs the WGSI server with the given provider
    """
    handler = PROVIDER_MAP[config.wsgi_provider]

    signal.signal(signal.SIGTERM, raise_ki_on_signal)

    handler(config, wsgi_application=functools.partial(restapi_wsgi, config, servlet))


def raise_ki_on_signal(signum: Any, frame: Any) -> None:
    """
    Raises a KeyboardInterrupt

    For use in signal handling
    """
    del signum
    del frame

    raise KeyboardInterrupt


def run_server_wsgiref(config: Config, wsgi_application: "WSGIApplication") -> None:
    """
    Runs the server using the wsgi reference application
    """
    logger = config.get_logger()

    import wsgiref.simple_server  # noqa: PLC0415

    dev_container = setup_mongo_container(config)

    server = wsgiref.simple_server.make_server(
        config.http_server_bind,
        config.http_server_port,
        wsgi_application,
    )

    with server as httpd:
        logger.info(
            "wsgiref listening on http://%s:%s/", config.http_server_bind, config.http_server_port
        )
        logger.info("CBC_PATH: %s", config.cbc_path)
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            logger.info("Caught keyboard interrupt")
            if dev_container:
                shutdown_mongo_container(config, dev_container)
            httpd.shutdown()


PROVIDER_MAP = {
    WsgiProvider.WSGIREF: run_server_wsgiref,
}



"""Cloud Function response module."""

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from ixoncdkingress.types import ContentType, ResponseCode

if TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse


class Response:
    """
    Represents an HTTP response, containing headers and a body
    """

    body: bytes | None
    content_type: ContentType | None
    cookie: dict[str, str]
    headers: list[tuple[str, str]]
    start_response: "StartResponse"
    status_code: ResponseCode

    def __init__(self, start_response: "StartResponse") -> None:
        self.body = None
        self.content_type = None
        self.cookie = {}
        self.headers = []
        self.start_response = start_response
        self.status_code = ResponseCode.OK

    def __call__(self, *args: Any, **kwargs: Any) -> Iterable[bytes]:
        del args
        del kwargs

        headers = [*self.headers, *[("Set-Cookie", f"{k}={v}") for (k, v) in self.cookie.items()]]

        if self.content_type:
            headers.insert(0, ("Content-Type", f"{self.content_type.value}; charset=utf-8"))

        self.start_response(self.status_code.status_line(), headers)

        if self.body:
            return [self.body]

        return []

    def set_body(self, body: bytes) -> None:
        """
        Sets the body of the response
        """
        self.body = body



"""
Utility functions.
"""

import json
from collections.abc import Mapping
from time import sleep
from typing import TYPE_CHECKING, Any, Optional, cast
from urllib.parse import parse_qs

from docker import errors, from_env
from pymongo import MongoClient

from ixoncdkingress.function.api_client import ApiClient, Query
from ixoncdkingress.function.context import FunctionContext, FunctionResource
from ixoncdkingress.types import FunctionArguments, FunctionLocation
from ixoncdkingress.webserver.config import Config

if TYPE_CHECKING:
    from docker.models.containers import Container

_DEV_MONGO_USERNAME = "root"
_DEV_MONGO_PASSWORD = "mongo-root"  # noqa: S105 # nosec
_DEV_MONGO_DEFAULT_COLLECTION = "cloud_function_pid"

_MONGO_TIMEOUT_MS: int = 30000


def read_qs_as_dict(in_put: bytes) -> dict[str, str]:
    """
    Reads and parses the URL query string, returns it as a dict with for each value only the first
    occurrence.
    """
    body = in_put.decode("utf-8")

    # parse_qs always returns non-empty lists!
    return {key: value[0] for (key, value) in parse_qs(body).items()}


def parse_function_location(in_put: str) -> FunctionLocation:
    """
    Parses the full function path into a module path and a function name.
    """

    function_path = in_put.split(".")

    function_name = function_path.pop()
    module_name = ".".join(function_path)

    if not function_path:
        return "functions", function_name

    return module_name, function_name


def _parse_context_values(context_items: dict[str, Any]) -> dict[str, str | FunctionResource]:
    """
    Parses expected context_items into FunctionResources
    and discards any unexpected context_items.
    """
    context_values = {}
    for context_name, context_resource in context_items.items():
        # context_values.apiApplication and .accessToken will be sent to the FunctionContext
        # but are unused, they are only used to create the ApiClient.
        if context_name not in [
            "user",
            "company",
            "agent",
            "asset",
            "apiApplication",
            "accessToken",
            "template",
        ]:
            continue

        if context_resource is None:
            continue

        # accessToken is not a FunctionResource because it doesn't have
        # a publicId, name and custom field.
        if context_name == "accessToken":
            context_values[context_name] = context_resource.get("headerValue")
            continue

        permissions = None
        if context_name in ["company", "agent", "asset"]:
            permissions = set()
            if "permissions" in context_resource:
                permissions.update(context_resource["permissions"])

        context_values[context_name] = FunctionResource(
            context_resource["publicId"],
            context_resource["name"],
            context_resource["custom"],
            permissions,
        )

    return context_values


def parse_json_input(
    cdk_config: Config, context_config: dict[str, str], body: str
) -> tuple[FunctionContext, FunctionLocation, FunctionArguments, str]:
    """
    Parses an application/json request body string into a context, function
    location and function arguments.
    """
    in_put = json.loads(body)

    function_location = parse_function_location(in_put.get("name", ""))

    # Get all context values, may be overwritten if a custom apiApplication is set
    # however this is then still needed for getting authentication values
    context_values = _parse_context_values(in_put.get("context", {}))

    context_config = context_config | in_put.get("config", {})

    api_client_kwargs = {}

    if (api_application_res := context_values.get("apiApplication")) and isinstance(
        api_application_res, FunctionResource
    ):
        api_client_kwargs["api_application"] = api_application_res.public_id
    if (api_company_res := context_values.get("company")) and isinstance(
        api_company_res, FunctionResource
    ):
        api_client_kwargs["api_company"] = api_company_res.public_id
    if (access_token := context_values.get("accessToken")) and isinstance(access_token, str):
        api_client_kwargs["authorization"] = access_token

    api_client = ApiClient(cdk_config.api_client_base_url, **api_client_kwargs)

    # Override the API application if it's set in context_config and not in production mode
    # and refetch custom properties of context_values for scoped custom properties.
    if not cdk_config.production_mode and (
        api_application_res := context_config.get("apiApplication")
    ):
        query: Query = {"fields": "custom,name"}
        api_client.set_custom_api_application(api_application_res)

        # Without an access token we don't requery the custom properties
        # as the request will fail anyways.
        if context_values.get("accessToken"):
            api_data = {}
            user_res = api_client.get("MyUser", query=query)
            api_data["user"] = user_res["data"]

            # Without a company we don't need to requery these properties
            # as the requests will fail anyways.
            if context_values.get("company"):
                company_res = api_client.get("MyCompany", query=query)
                api_data["company"] = company_res["data"]

                if (agent := context_values.get("agent")) and isinstance(agent, FunctionResource):
                    agent_res = api_client.get(
                        "Agent",
                        {"publicId": agent.public_id},
                        query=query,
                    )
                    api_data["agent"] = agent_res["data"]

                if (asset := context_values.get("asset")) and isinstance(asset, FunctionResource):
                    asset_res = api_client.get(
                        "Asset",
                        {"publicId": asset.public_id},
                        query=query,
                    )
                    api_data["asset"] = asset_res["data"]

            # Overwrite existing context_values
            context_values = _parse_context_values(api_data)

    context_internal_config = in_put.get("internalConfig", {})

    context = FunctionContext(
        context_config,
        api_client,
        create_mongo_client(
            cdk_config,
            context_internal_config,
        ),
        user=cast("FunctionResource", context_values.get("user", None)),
        company=cast("FunctionResource", context_values.get("company", None)),
        asset=cast("FunctionResource", context_values.get("asset", None)),
        agent=cast("FunctionResource", context_values.get("agent", None)),
        template=cast("FunctionResource", context_values.get("template", None)),
        document_db_collection_name=get_document_db_collection_name(context_internal_config),
    )

    function_arguments = in_put.get("arguments", {})
    created_on = in_put.get("created_on", "")

    return context, function_location, function_arguments, created_on


def get_document_db_collection_name(internal_config: Mapping[str, object] | None = None) -> str:
    """
    Returns the Document DB collection name to use
    """
    if internal_config is None:
        return _DEV_MONGO_DEFAULT_COLLECTION

    return cast("str", internal_config.get("publicId", _DEV_MONGO_DEFAULT_COLLECTION))


def create_mongo_client(
    cdk_config: Config,
    internal_config: Mapping[str, object],
) -> MongoClient[Mapping[str, object]] | None:
    """
    Create a MongoDB client based on the given internal config & ixoncdkingress config.

    If the ixoncdkingress is running in development-mode, and there is no config
    passed via the internal_config, we default to connecting to the mongodb server
    that is started as part of the ixoncdkingress development workspace.
    """
    if (db_config := internal_config.get("dbConfig")) and isinstance(db_config, dict):
        return MongoClient(
            db_config["uri"],
            username=db_config["username"],
            password=db_config["password"],
            timeoutMS=_MONGO_TIMEOUT_MS,
            tls=True,
        )

    if not cdk_config.production_mode:
        return MongoClient(
            f"mongodb://localhost:{cdk_config.document_db_port}/",
            username=_DEV_MONGO_USERNAME,
            password=_DEV_MONGO_PASSWORD,
            timeoutMS=_MONGO_TIMEOUT_MS,
            tls=False,
        )

    # Production but no/invalid db config
    return None


def setup_mongo_container(config: Config) -> Optional["Container"]:
    """
    Sets up a development MongoDB when not in production mode.

    Will create a docker container named: `ixon-cloud-function-mongo`. The mongodb will be
    accessable at the configured port with user: `root` and password: `mongo-root`

    Returns the created container so that it can be removed on exit.
    """
    if config.production_mode:
        return None

    logger = config.get_logger()
    logger.info("Starting DocumentDB server")

    try:
        client = from_env()

        container: Container
        try:
            container = client.containers.get("ixon-cloud-function-mongo")
            container.remove(force=True)
        except errors.NotFound:
            pass

        container = client.containers.run(
            "mongo:6.0.6-jammy",
            name="ixon-cloud-function-mongo",
            ports={"27017/tcp": config.document_db_port},
            detach=True,
        )

        # Ensure database is running and user created. has to be variable because of possible
        # slow development computers of users
        retries, exit_code = 0, 1
        max_retries = 10
        while retries < max_retries and exit_code != 0:
            sleep(1 * retries)
            retries += 1
            exit_code = container.exec_run(
                'mongosh --eval \'db.getSiblingDB("admin").createUser({\n'
                f'   user: "{_DEV_MONGO_USERNAME}",\n'
                f'   pwd: "{_DEV_MONGO_PASSWORD}",\n'
                '   roles: [ { role: "userAdminAnyDatabase", db: "admin" } ]\n'
                "})'\n"
            ).exit_code

        logger.info("DocumentDB server started successfully")

        return container
    except errors.DockerException as error:
        logger.error(
            "\
            Exception with setting up DocumentDB. \
            Please make sure you have docker installed on your machine \
            and port %s is unused or configure a different port \
            by setting DOCUMENT_DB_PORT to an available port\
        ",
            config.document_db_port,
        )
        raise error


def shutdown_mongo_container(config: Config, container: "Container") -> None:
    """
    Tears down and removes the earlier created DocumentDB container
    """
    logger = config.get_logger()

    logger.info("Shutting down DocumentDB server")
    container.remove(force=True)



"""
HTML form generator
"""

from typing import TYPE_CHECKING
from urllib.parse import parse_qs

from ixoncdkingress.function.api_client import ApiClient
from ixoncdkingress.function.context import FunctionContext, FunctionResource
from ixoncdkingress.types import FunctionArguments, FunctionLocation
from ixoncdkingress.webserver.config import Config
from ixoncdkingress.webserver.utils import (
    create_mongo_client,
    get_document_db_collection_name,
    parse_function_location,
)

try:
    from dominate.document import document
    from dominate.tags import (
        div,
        form,
        h1,
        h3,
        input_,
        label,
        table,
        tbody,
        td,
        textarea,
        th,
        thead,
        tr,
    )

    HAS_DOMINATE = True  # pragma: no cover
except ImportError:  # pragma: no cover
    HAS_DOMINATE = False

if TYPE_CHECKING:
    from dominate import tags


def generate_form(pre_fill: dict[str, str], form_output: str | None = None) -> str:
    """
    Generates and renders the CDK Ingress HTML form
    """
    return _generate_form_document(pre_fill, form_output).render()


def _generate_form_document(pre_fill: dict[str, str], form_output: str | None = None) -> "document":
    """
    Generates the CDK Ingress HTML form document
    """
    if not HAS_DOMINATE:
        raise Exception("Requires dominate package. pip install dominate")

    doc = document(title="IXON CDK Ingress")

    with doc:
        h1("IXON CDK Ingress")
        with form(method="post"):
            with div(class_name="function"):
                label("Function", _for="function")
                input_(
                    type="text",
                    id="function",
                    name="function",
                    size="32",
                    required="true",
                    value=pre_fill.get("function"),
                )
            with div(class_name="arguments"):
                h3("Function input")
                _custom_properties(pre_fill, prefix="function_args__")

            with div(class_name="resource"):
                h3("User")
                _resource(pre_fill, prefix="user__", show_permission=False)
            with div(class_name="resource"):
                h3("Company")
                _resource(pre_fill, prefix="company__", show_permission=True)
            with div(class_name="resource"):
                h3("Asset")
                _resource(pre_fill, prefix="asset__", show_permission=True)
            with div(class_name="resource"):
                h3("Agent")
                _resource(pre_fill, prefix="agent__", show_permission=True)

            input_(type="submit", value="Run function")
            input_(type="reset", value="Clear form")

        h3("Function output")
        textarea(form_output or "Function output goes here", rows="10", cols="80")

    return doc


def _custom_properties(pre_fill: dict[str, str], prefix: str) -> "tags.html_tag":
    """
    Generates a form fragment for custom properties.
    """
    table_tag = table()

    with table_tag:
        with thead():
            th("Key")
            th("Value")
        with tbody():
            for i in range(9):
                with tr():
                    td(
                        input_(
                            type="text",
                            name=f"{prefix}key{i}",
                            id=f"{prefix}key{i}",
                            size="16",
                            value=pre_fill.get(f"{prefix}key{i}"),
                        )
                    )
                    td(
                        input_(
                            type="text",
                            name=f"{prefix}value{i}",
                            id=f"{prefix}value{i}",
                            size="32",
                            value=pre_fill.get(f"{prefix}value{i}"),
                        )
                    )

    return table_tag


def _permissions(pre_fill: dict[str, str], prefix: str) -> "tags.html_tag":
    """
    Generates a form fragment for permissions.
    """
    table_tag = table()

    with table_tag, tbody():
        for i in range(9):
            with tr():
                td(
                    input_(
                        type="text",
                        name=f"{prefix}value{i}",
                        id=f"{prefix}value{i}",
                        size="32",
                        value=pre_fill.get(f"{prefix}value{i}"),
                    )
                )

    return table_tag


def _resource(pre_fill: dict[str, str], prefix: str, show_permission: bool) -> "tags.html_tag":
    """
    Generates a form fragment for a context resource.
    """
    table_tag = table()

    with table_tag, tbody():
        with tr():
            td(label("PublicID", _for=f"{prefix}public_id"))
            td(
                input_(
                    type="text",
                    id=f"{prefix}public_id",
                    name=f"{prefix}public_id",
                    size="12",
                    value=pre_fill.get(f"{prefix}public_id"),
                )
            )
        with tr():
            td(label("Name", _for=f"{prefix}name"))
            td(
                input_(
                    type="text",
                    id=f"{prefix}name",
                    name=f"{prefix}name",
                    size="32",
                    value=pre_fill.get(f"{prefix}name"),
                )
            )
        with tr():
            td("Custom properties")
            td(_custom_properties(pre_fill, f"{prefix}custom_properties__"))

        if show_permission:
            td("Permissions")
            td(_permissions(pre_fill, f"{prefix}permissions__"))

    return table_tag


def parse_form_input(
    config: Config, context_config: dict[str, str], body: str
) -> tuple[FunctionContext, FunctionLocation, FunctionArguments, str]:
    """
    Parses an application/x-www-form-urlencoded request body string into a context, function
    location and function arguments.
    """

    in_put = parse_qs(body)

    context = FunctionContext(
        context_config,
        api_client=ApiClient(config.api_client_base_url),
        mongo_client=create_mongo_client(config, {}),
        user=_parse_resource("user__", in_put, False),
        company=_parse_resource("company__", in_put, True),
        asset=_parse_resource("asset__", in_put, True),
        agent=_parse_resource("agent__", in_put, True),
        template=_parse_resource("template__", in_put, False),
        document_db_collection_name=get_document_db_collection_name(),
    )

    function_location = parse_function_location(in_put.get("function", [""])[0])
    function_arguments = _parse_function_args(in_put)

    return context, function_location, function_arguments, ""


def _parse_function_args(in_put: dict[str, list[str]]) -> FunctionArguments:
    function_arguments = {}
    for in_put_key, in_put_value in in_put.items():
        if not in_put_key.startswith("function_args__key"):
            continue

        key_argument_value = f"function_args__value{in_put_key[18:]}"
        if key_argument_value not in in_put:
            continue

        function_arguments[in_put_value[0]] = in_put[key_argument_value][0]

    return function_arguments


def _parse_resource(
    prefix: str, in_put: dict[str, list[str]], has_permissions: bool
) -> FunctionResource | None:
    """
    Reads a resource description from the in_put and returns it as a FunctionResource.
    """
    props = {}
    permissions: set[str] | None = set() if has_permissions else None
    if not in_put.get(f"{prefix}public_id") or not in_put.get(f"{prefix}name"):
        return None

    custom_property_prefix = f"{prefix}custom_properties__"
    permission_prefix = f"{prefix}permissions__value"
    for in_put_key, in_put_value in in_put.items():
        # Process custom property
        if in_put_key.startswith(f"{custom_property_prefix}key"):
            custom_property_index = in_put_key[len(f"{custom_property_prefix}key") :]
            key_custom_property_value = f"{custom_property_prefix}value{custom_property_index}"
            if key_custom_property_value not in in_put:
                continue

            props[in_put_value[0]] = in_put[key_custom_property_value][0]
            continue

        # Process permission
        if has_permissions and in_put_key.startswith(permission_prefix):
            assert isinstance(permissions, set)  # type hint
            permissions.add(in_put_value[0])

    return FunctionResource(
        in_put[f"{prefix}public_id"][0], in_put[f"{prefix}name"][0], props, permissions
    )



"""
Contains code dealing with API clients
"""

import json
from collections.abc import MutableMapping
from typing import Any

import requests

Query = MutableMapping[str, None | str | list[str]]

ClientRequest = dict[str, Any] | list[dict[str, Any]] | bytes

ServerResponse = dict[str, Any]


class ApiClient:
    """
    class for a API client
    """

    _base_url: str
    """
    The base URL (entry point) of the API server
    """

    _default_headers: dict[str, str]
    """
    The headers to send along with every request
    """

    _session: requests.Session
    """
    Requests session to run requests with
    """

    _discovery: dict[str, str] | None
    """
    Discovery dictionary
    """

    def __init__(
        self,
        base_url: str,
        api_application: str | None = None,
        api_company: str | None = None,
        authorization: str | None = None,
    ) -> None:
        self._base_url = base_url

        self._default_headers = {
            "Api-Version": "2",
        }

        if api_application:
            self._default_headers["Api-Application"] = api_application
        if api_company:
            self._default_headers["Api-Company"] = api_company
        if authorization:
            self._default_headers["Authorization"] = authorization

        self._session = requests.session()

        self._discovery = None

    def _check_discovery(self, verify: bool = True) -> None:
        """
        Checks whether discovery has been run, and, if not, runs it.

        Set verify on True when the certificate shouldn't be checked, which is the case when
        the ApiClient has to talk to the envdev ixapi-server.

        We don't expire the discovery as a new ApiClient instance is
        created for every Cloud Function call.
        """
        if self._discovery is not None:
            return

        discovery_headers = {"Api-Version": self._default_headers["Api-Version"]}
        response = self._session.get(self._base_url, headers=discovery_headers, verify=verify)
        assert response.ok, "Api discovery failed"
        data = response.json()
        assert "Discovery" == data["type"]
        self._discovery = {discovery["rel"]: discovery["href"] for discovery in data["data"]}

    def set_custom_api_application(self, api_application: str) -> None:
        """
        Modifies the Api-Application header used for API server requests.
        """
        self._default_headers["Api-Application"] = api_application

    def get(
        self,
        url_name: str,
        url_args: dict[str, str] | None = None,
        query: Query | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> ServerResponse:
        """
        Performs a GET request on the API server, on the URL with the given
        name

        The URL names are found in the discovery
        """
        return self._request(
            "GET", url_name, url_args=url_args, query=query, headers=headers, **kwargs
        )

    def patch(
        self,
        url_name: str,
        data: ClientRequest,
        url_args: dict[str, str] | None = None,
        query: Query | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> ServerResponse:
        """
        Performs a PATCH request on the API server, on the URL with the given
        name

        The URL names are found in the discovery
        """
        return self._request(
            "PATCH", url_name, url_args=url_args, query=query, data=data, headers=headers, **kwargs
        )

    def post(
        self,
        url_name: str,
        data: ClientRequest,
        url_args: dict[str, str] | None = None,
        query: Query | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> ServerResponse:
        """
        Performs a POST request on the API server, on the URL with the given
        name

        The URL names are found in the discovery
        """
        return self._request(
            "POST", url_name, url_args=url_args, query=query, data=data, headers=headers, **kwargs
        )

    def options(
        self,
        url_name: str,
        url_args: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> ServerResponse:
        """
        Performs an OPTIONS request on the API server, on the URL with the given
        name

        The URL names are found in the discovery

        When using a browser, these are done automatically. When not using a browser,
        these are not required
        """
        return self._request("OPTIONS", url_name, url_args=url_args, headers=headers, **kwargs)

    def put(
        self,
        url_name: str,
        url_args: dict[str, str] | None = None,
        data: ClientRequest | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """
        Performs a PUT request on the API server, on the URL with the given
        name

        The URL names are found in the discovery
        """
        return self._request(
            "PUT", url_name, url_args=url_args, data=data, headers=headers, **kwargs
        )

    def delete(
        self,
        url_name: str,
        url_args: dict[str, str] | None = None,
        data: ClientRequest | None = None,
        query: Query | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> ServerResponse:
        """
        Performs a DELETE request on the API server, on the URL with the given
        name

        The URL names are found in the discovery
        """
        return self._request(
            "DELETE", url_name, url_args=url_args, query=query, data=data, headers=headers, **kwargs
        )

    def _request(
        self,
        method: str,
        url_name: str,
        url_args: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        query: Query | None = None,
        data: ClientRequest | None = None,
        **kwargs: Any,
    ) -> ServerResponse:
        """
        Sends a request to the API server
        """
        if url_args is None:
            url_args = {}

        url = self._get_url(url_name, url_args, kwargs.get("verify", True))

        if headers is None:
            headers = {}

        data_json: bytes | None
        if data is None or isinstance(data, bytes):
            data_json = data
        else:
            data_json = json.dumps(data).encode("UTF-8")

        response = self._session.request(
            method,
            url,
            params=query,
            headers={
                **self._default_headers,
                **headers,
            },
            data=data_json,
            **kwargs,
        )
        try:
            return response.json()  # type: ignore[no-any-return]
        except json.decoder.JSONDecodeError:
            return {
                "type": "Error",
                "data": [
                    {
                        "message": response.text,
                    }
                ],
                "status": "error",
            }

    def _get_url(self, url_name: str, url_args: dict[str, str], verify: bool = True) -> str:
        """
        Returns an URL by name from the discovery and apply the URL arguments
        """
        self._check_discovery(verify=verify)
        assert self._discovery is not None  # type hint

        return self._discovery[url_name].format(**url_args)

    def __repr__(self) -> str:
        return f"<ApiClient base_url={self._base_url}, default_headers={self._default_headers},>"



"""
Time-series client & related types
"""

from collections.abc import Mapping, Sequence
from typing import Any, TypeAlias

from pymongo.command_cursor import CommandCursor

from ixoncdkingress.function._collection import Collection

_EntryType: TypeAlias = Mapping[str, object]


class TimeSeriesClient(Collection[_EntryType]):
    """
    Client for a Time Series collection.
    """

    def aggregate(self, pipeline: Sequence[Mapping[str, Any]]) -> CommandCursor[_EntryType]:
        """
        Run the given aggregation pipeline against the collection
        """
        return self._collection.aggregate(pipeline)



"""
Module containing utility functions
"""

import traceback

from ixoncdkingress.types import ResponseCode
from ixoncdkingress.webserver.response import Response


def handle_exception(exception: BaseException, response: Response) -> None:
    """
    Shows the error with traceback in the console and sets the response status code to 500.
    """
    traceback.print_exception(None, exception, exception.__traceback__)

    response.status_code = ResponseCode.INTERNAL_ERROR



"""
Functions for finding and calling Cloud Functions.
"""

import importlib
import inspect
import os
import pkgutil
import sys
import warnings
from functools import lru_cache
from typing import Any

from pymongo.errors import OperationFailure

from ixoncdkingress.function.context import FunctionContext
from ixoncdkingress.types import (
    AppDescription,
    AppDescriptionData,
    AppDescriptionResponse,
    FunctionArguments,
    FunctionLocation,
    ListToolsIssuesResponse,
    ListToolsResponse,
    ModuleLoadIssue,
    Tool,
)
from ixoncdkingress.utils import handle_exception
from ixoncdkingress.webserver.config import Config
from ixoncdkingress.webserver.response import Response


def call_function(
    config: Config,
    context: FunctionContext,
    function_location: FunctionLocation,
    function_kwargs: FunctionArguments,
    response: Response,
) -> Any:
    """
    Finds, loads and calls the function specified in the body. The content_type specifies the
    format of the body.
    """

    # Get the specified module
    if function_location[1] == "list-agent-tools":
        return list_agent_tools(config.cbc_path)

    if function_location[1] == "list-agent-tools-issues":
        return list_agent_tools_issues(config.cbc_path)

    if function_location[1] == "app-description":
        return get_app_description(config.cbc_path)

    sys.path.insert(0, config.cbc_path)
    module = importlib.import_module(function_location[0])
    if config.production_mode is False:
        try:
            module = importlib.reload(module)
        except ImportError as error:
            handle_exception(error, response)
            return None
    del sys.path[0]

    # Get the specified function
    if hasattr(module, function_location[1]):
        function = getattr(module, function_location[1])

        # Check if it's exposed
        if getattr(function, "exposed", False):
            try:
                return function(context, **get_function_args(function, function_kwargs))
            except OperationFailure as ex:
                if ex.details and ex.details.get("codeName"):
                    return f"Document DB operation failure: {ex.details['codeName']}"

                return "Document DB operation failure"

    return f"Not found, function '{'.'.join(function_location)}' does not exist or is not exposed"


def get_function_args(function: Any, function_kwargs: dict[Any, Any]) -> dict[Any, Any]:
    """
    Get the arguments of the function.

    Discards arguments that the function does not take.
    """
    params = inspect.signature(function).parameters

    # Check if the function takes a variable number of keyword arguments
    if any(key.kind is inspect.Parameter.VAR_KEYWORD for key in params.values()):
        return function_kwargs

    input_model = getattr(function, "input_model", None)

    if input_model is not None:
        # Create pydantic models for the function arguments
        args = dict(vars(input_model(**function_kwargs)))
    else:
        # Discard all arguments that the function can not take
        args = {key: value for key, value in function_kwargs.items() if key in params}

    return args


def _get_tools(cbc_path: str) -> tuple[list[Tool], list[ModuleLoadIssue]]:
    sys.path.insert(0, cbc_path)
    tools: list[Tool] = []
    issues: list[ModuleLoadIssue] = []

    # Discover all .py files under the cbc_path (package and submodules)
    for module_name in _walk_modules(cbc_path):
        try:
            module = importlib.import_module(module_name)
        except Exception as e:  # noqa: BLE001
            warnings.warn(f"Failed to import module {module_name}, reason: {e!r}", stacklevel=2)
            issues.append(
                ModuleLoadIssue(
                    type="import_error",
                    module_name=module_name,
                    error=str(e),
                )
            )
            continue

        for obj in module.__dict__.values():
            # Only consider functions or callables
            if (
                callable(obj)
                and getattr(obj, "exposed_to_agent", False)
                and hasattr(obj, "tool_definition")
            ):
                tool_definition = obj.tool_definition
                tool_definition.module_name = module_name
                tools.append(tool_definition)

    del sys.path[0]

    return tools, issues


@lru_cache(maxsize=1)
def list_agent_tools(cbc_path: str) -> ListToolsResponse:
    """
    List the tools available to the agent.
    """

    tools, _ = _get_tools(cbc_path)

    return ListToolsResponse(
        result="success",
        data=[tool.to_dict() for tool in sorted(tools, key=lambda x: x.name) if tool.is_ready],
    )


@lru_cache(maxsize=1)
def list_agent_tools_issues(cbc_path: str) -> ListToolsIssuesResponse:
    """
    List the the tools that are not available to the agent, and the reasons why
    they are not available.
    """

    tools, issues = _get_tools(cbc_path)

    return ListToolsIssuesResponse(
        result="success",
        data=issues + [tool.to_dict() for tool in tools if not tool.is_ready],
    )


def _walk_modules(package_path: str, prefix: str = "") -> list[str]:
    """Recursively discover all modules and submodules under package_path."""
    modules = []
    for _, modname, ispkg in pkgutil.iter_modules([package_path]):
        full_name = prefix + modname if prefix == "" else prefix + "." + modname
        modules.append(full_name)
        if ispkg:
            sub_path = os.path.join(package_path, modname)
            modules.extend(_walk_modules(sub_path, full_name))
    return modules


@lru_cache(maxsize=1)
def get_app_description(cbc_path: str) -> AppDescriptionResponse:
    _ = _get_tools(cbc_path)  # load all modules to set the app description

    try:
        app_description = AppDescription.get()
    except ValueError:
        app_description = None

    return AppDescriptionResponse(
        result="success",
        data=AppDescriptionData(
            description=app_description.description if app_description else None,
            shortName=app_description.short_name if app_description else None,
        ),
    )



"""Cloud Function request module."""

import json
from http.cookies import SimpleCookie
from typing import TYPE_CHECKING, Any
from urllib.parse import parse_qs

from ixoncdkingress.types import ContentType, RequestMethod
from ixoncdkingress.webserver.config import Config

if TYPE_CHECKING:
    from _typeshed.wsgi import WSGIEnvironment


class Request:
    """
    Represents an HTTP request, containing headers and optionally a body
    """

    body: dict[str, Any] | None
    config: Config
    content_type: ContentType | None
    authorization: str | None
    cookies: SimpleCookie | None
    request_body: bytes
    request_method: RequestMethod

    def __init__(self, config: Config, environ: "WSGIEnvironment") -> None:
        self.config = config
        self.request_method = RequestMethod(environ["REQUEST_METHOD"])
        self.body = None

        if cookies_env := environ.get("HTTP_COOKIE"):
            self.cookies = SimpleCookie(cookies_env)
        else:
            self.cookies = None

        try:
            request_body_size = int(environ.get("CONTENT_LENGTH", 0))
            self.request_body = environ["wsgi.input"].read(request_body_size)
        except ValueError:
            self.request_body = b""

        try:
            self.content_type = ContentType(environ["CONTENT_TYPE"])
        except (KeyError, ValueError):
            self.content_type = None

        try:
            self.authorization = environ["HTTP_AUTHORIZATION"]
        except KeyError:
            self.authorization = None

    def parse_body(self) -> None:
        """
        Parses the request body according to the `content_type`.

        Must only be called once.
        """
        assert self.body is None

        if ContentType.FORM == self.content_type:
            self.body = parse_qs(self.request_body.decode("utf-8"))
        elif ContentType.JSON == self.content_type:
            self.body = json.loads(self.request_body)
        else:
            raise NotImplementedError



"""
Document DB client & related types
"""

import warnings
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import Any, TypeAlias

from pymongo.results import (
    DeleteResult,
    InsertManyResult,
    InsertOneResult,
    UpdateResult,
)

from ixoncdkingress.function._collection import Collection

DocumentType: TypeAlias = Mapping[str, Any]


class DocumentDBClient(Collection[DocumentType]):  # pragma: no-cov
    """
    Client for a Document DB collection.
    """

    def insert_one(self, document: DocumentType) -> InsertOneResult:
        """
        Insert a single document into the store.
        """
        return self._collection.insert_one(document)

    def insert_many(self, documents: Iterable[DocumentType]) -> InsertManyResult:
        """
        Insert an iterable of documents into of store.
        """
        return self._collection.insert_many(documents)

    def update_one(self, filter_map: Mapping[str, Any], update: Mapping[str, Any]) -> UpdateResult:
        """
        Update a single document matching the filter.
        """
        return self._collection.update_one(filter_map, update)

    def update_many(self, filter_map: Mapping[str, Any], update: Mapping[str, Any]) -> UpdateResult:
        """
        Update one or more documents that match the filter.
        """
        return self._collection.update_many(filter_map, update)

    def delete_one(self, filter_map: Mapping[str, Any]) -> DeleteResult:
        """
        Delete a single document matching the filter.
        """
        return self._collection.delete_one(filter_map)

    def delete_many(self, filter_map: Mapping[str, Any]) -> DeleteResult:
        """
        Delete one or more documents matching the filter.
        """
        return self._collection.delete_many(filter_map)


@dataclass(frozen=True, init=False)
class DocumentDBAuthentication:
    """
    DEPRECATED: Do not use.
    """

    username: str
    password: str

    def __init__(self, username: str, password: str):
        warnings.warn(DeprecationWarning("Deprecated: do not use"))

        # Need to use __setattr__ on object since dataclass is frozen
        object.__setattr__(self, "username", username)
        object.__setattr__(self, "password", password)


def __getattr__(name: str) -> object:  # pragma: no cover
    """
    Handle accessing deprecated module attributes.
    """
    if name == "TIMEOUT":
        warnings.warn(DeprecationWarning("Deprecated: do not use"), stacklevel=1)
        return 3000  # Return the last value TIMEOUT had before deprecation

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""
The __main__ module lets the user start the ingress from the command line
"""

import sys

from ixoncdkingress.webserver import wsgi
from ixoncdkingress.webserver.config import get_config
from ixoncdkingress.webserver.servlet import Servlet


def main(argv: list[str]) -> int:
    """
    Parses the arguments and runs the provider
    """
    del argv  # Configuration goes via environment

    config = get_config()
    servlet = Servlet(config)

    try:
        wsgi.run_server(config, servlet)
        return 0
    except KeyboardInterrupt:
        ...

    return 0


if "__main__" == __name__:  # pragma: no cover
    sys.exit(main(sys.argv[1:]))
